#!/bin/sh
#
# DoudouLinux sound, function library
#
# Requires alsa-utils, hal.
#
# Author: JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

#TODO: use dbus to raise user's apps like vmpk

# constants
ASOUNDCONF=/etc/asound.conf
DDLSOUNDSERV=/etc/init.d/doudoulinux-sound
TIMIDITYCONF=/etc/default/timidity
TIMIDITYSERVICE=/etc/init.d/timidity

#################
start_vmpk() {
	# starts VMPK if not already started
	#
	# start_vmpk
	
	# don't do anything if VMPK is not installed or already running
	if ( ! [ -x "$(which vmpk)" ] || aconnect -i | grep VMPK >/dev/null ); then
		return 0
	fi
	
	# get X display or abort
	# NB: cannot be put in a function apart
	USERNAME=$(who | grep 'tty.*(:[0-9])' | head -n 1 | grep -o '^[^ ]*')
	if [ -n "$USERNAME" ]; then
		USERHOME=$(grep "^$USERNAME:" /etc/passwd | cut -d : -f 6)
		export DISPLAY=$(who | grep 'tty.*(:[0-9])' | head -n 1 | cut -d '(' -f 2 | grep -o ':[0:9]')
		export XAUTHORITY=$USERHOME/.Xauthority
	else
		# no user logged in
		return 1
	fi
	
	# compute VMPK window geometry then start VMPK, full width
	WIDTH=$(xrandr | grep '*' | grep -o '[0-9]*' | head -n 1)
	HEIGHT=$(( (10 * $WIDTH) / 39 ))
	Y=$(xrandr | grep '*' | grep -o '[0-9]*' | head -n 2 | tail -n 1)
	Y=$(( $Y - $HEIGHT ))
	su -c "vmpk -geometry $WIDTH\x$HEIGHT+0+$Y" $USERNAME &

	# wait for the VMPK MIDI input to show up
	TIMEOUT=10
	while [ $TIMEOUT -ge 0 ]; do
		TIMEOUT=$(( $TIMEOUT - 1 ))
		sleep 1
		if aconnect -o | grep VMPK >/dev/null; then
			return 0
		fi
	done
	
	# failed to start
	return 1
}

#################
connect_midikbd() {
	# connects an input MIDI keyboard to Timidity and VMPK
	#
	# connect_midikbd DEVNAME
	DEVNAME=$(echo $1 | sed 's/_/ /g')		# replace '_' with ' '
	
	# get Timidity MIDI id.
	TIMIDITYID=$(aconnect -o | grep ^client | grep TiMidity)
	if [ -z "$TIMIDITYID" ]; then
		# no timidity running
		return 1
	fi
	
	# scan MIDI inputs
	aconnect -i | grep ^client | while read MIDIOUT; do
		MIDIOUT=$(echo $MIDIOUT | cut -d "'" -f 2)
		if echo $DEVNAME | grep "$MIDIOUT" >/dev/null; then
			if start_vmpk; then
				# connect keyboard to VMPK which should activate MIDI through
				echo "[doudoulinux] connecting '$MIDIOUT:0' to 'VMPK Input:0'"
				aconnect "$MIDIOUT:0" 'VMPK Input:0'
			else
				# connect keyboard directly to timidity
				echo "[doudoulinux] connecting '$MIDIOUT:0' to 'TiMidity:0'"
				aconnect "$MIDIOUT:0" TiMidity:0
			fi
		fi
	done
	return 0
}

#################
timidity_connected() {
	# collects the list of MIDI outputs connected to Timidity inputs
	#
	# timidity_connected
	
	# scan all the MIDI connections to find the Timidity ones
	CONNECTLIST=
	RECORD=0
	PORT=
	aconnect -ol | while read LINE; do
		# wait for the line "client 128: 'TiMidity' [type=user]"
		if [ $RECORD -eq 0 ]; then
			if ! echo $LINE | grep '^client ' >/dev/null; then
				continue
			fi
			if ! echo $LINE | grep 'TiMidity' >/dev/null; then
				continue
			fi
			RECORD=1
			continue
		fi
		
		# stop if a new client is encountered
		if echo $LINE | grep '^client ' >/dev/null; then
			RECORD=0
			continue
		fi
		
		# record connected clients
		if echo $LINE | grep '^[0-9] ' >/dev/null; then
			PORT=$(echo $LINE | sed -r 's/^[ ]*([0-9]) .*/\1/')
			continue
		fi
		if echo $LINE | grep '^Connected From: ' >/dev/null; then
			DEST=$(echo $LINE | sed -r 's/Connected From: ([0-9]+:[0-9]).*/\1/')
			echo $PORT/$DEST
		fi
	done
}

#################
timidity_reconnect() {
	# reconnects ports that were previously connected, before Timidity restart
	#
	# timidity_reconnect PAIR [PAIR…]

	# wait for the Timidity MIDI input to show up
	TIMEOUT=10
	while [ $TIMEOUT -ge 0 ]; do
		TIMEOUT=$(( $TIMEOUT - 1 ))
		sleep 1
		if aconnect -o | grep TiMidity >/dev/null; then
			break
		fi
	done
	
	# get the Timidity MIDI device number
	DEVICE=$(aconnect -o | grep '^client ' | grep TiMidity | sed -r 's/client ([0-9]+).*/\1/')
	if [ $? -ne 0 ]; then
		echo "[doudoulinux] cannot find a Timidity MIDI port"
		return 1
	fi
	
	# scan pairs of connections
	echo "[doudoulinux] reconnecting Timidity $@"
	for PAIR in $@; do
		PORT=$(echo $PAIR | cut -d '/' -f 1)
		DEST=$(echo $PAIR | cut -d '/' -f 2)
		aconnect $DEST $DEVICE:$PORT
	done
}

#################
restart_timidity() {
	# restarts Timidity as a service or as a user application
	# NB: if we are not root, we cannot restart the service
	#
	# restart_timidity
	
	# nothing to do if timidity is not running
	if [ -z "$(ps -C timidity -o user=)" ]; then
		return
	fi
	
	# list MIDI ports connected to Timidity
	CONNECTLIST=$(timidity_connected)
	
	# timidity started as a service?
	CURRENTUSER=$(whoami)
	if [ -f $TIMIDITYSERVICE ]; then
		TIM_ALSASEQ="true"
		if [ -f $TIMIDITYCONF ]; then
			# the configuration file may disallow running as a service
			. $TIMIDITYCONF
		fi
		if [ "$TIM_ALSASEQ" = "true" ]; then
			if [ "$CURRENTUSER" = "root" ]; then
				$TIMIDITYSERVICE restart
			fi
			timidity_reconnect $CONNECTLIST
			return
		fi
	fi
	
	# restart timidity as a user process
	ps -C timidity -o user=,pid=,cmd= | while read LINE; do
		# scan all users processes to kill then restart timidity ones
		USERNAME=$(echo $LINE | grep -o '^[^ ]*')
		if ( [ "$USERNAME" = "$CURRENTUSER" ] || [ "$CURRENTUSER" = "root" ] ); then
			CMD=$(echo $LINE | sed 's/^'$USERNAME'[ ]*[0-9]*[ ]*//')
			PID=$(echo $LINE | sed 's/^'$USERNAME'[ ]*//' | grep -o '^[0-9]*')
			echo "[doudoulinux] killing $PID"
			kill $PID
			echo "[doudoulinux] launching '$CMD' as user '$USERNAME'"
			if [ $USERNAME = $CURRENTUSER ]; then
				eval $CMD
			else
				su -c "$CMD" $USERNAME
			fi
		fi
		timidity_reconnect $CONNECTLIST
	done
}

#################
set_audiocard() {
	# changes the default ALSA card but preserves the audio card previously set
	#
	# set_audiocard --keep CARDNBR
	# set_audiocard CARDNBR CARDNAME
	
	# input args
	if [ "$1" = "--keep" ]; then
		KEEP=1
		CARDNBR=$2
		CARDNAME=
	else
		KEEP=0
		CARDNBR=$1
		CARDNAME=$2
	fi
	
	# try to preserve the name of the card previously set
	if [ $KEEP -eq 1 ]; then
		PREVIOUS='# no card set previously'
		if [ -f $ASOUNDCONF ]; then
			CARDNAME=$(grep '^# previously set:' $ASOUNDCONF | sed 's/# previously set: //')
			if [ $? -eq 0 ]; then
				PREVIOUS="# previously set: $CARDNAME"
			fi
		fi
	else
		PREVIOUS="# previously set: $CARDNAME"
	fi
	
	# change of audio card
	cat >$ASOUNDCONF <<EOF
$PREVIOUS
defaults.ctl.card $CARDNBR
defaults.pcm.card $CARDNBR
defaults.pcm.device 0
defaults.pcm.subdevice 0
EOF
}

#################
init_audiocard() {
	# inits the configuration file /etc/asound.conf
	# used at system boot or to restore the default card
	#
	# init_audiocard
	
	# initializations
	SELECTCARD=$(aplay -l | grep '^card' | head -n 1)
	if ! [ -f $ASOUNDCONF ]; then
		touch $ASOUNDCONF
	fi
	
	# no card → bye
	if [ -z "$SELECTCARD" ]; then
		return 1
	fi
	
	# search for a card previously set
	if grep '# previously set:' $ASOUNDCONF >/dev/null; then
		CARDNAME=$(grep '# previously set:' $ASOUNDCONF | sed 's/# previously set:[ ]*//')
		echo "doudoulinux-sound: previous sound card = $CARDNAME"
		PREVIOUSCARD=$(aplay -l | grep '^card' | grep "\[$CARDNAME\]$")
		if [ -n "$PREVIOUSCARD" ]; then
			# previously selected card is still in current hardware configuration
			SELECTCARD=$PREVIOUSCARD
		fi
	fi
	
	# compute card number and name
	CARDNBR=$(echo "$SELECTCARD" | head -n 1 | sed 's/^card \([0-9]\).*/\1/')
	CARDNAME=$(echo "$SELECTCARD" | sed 's/.*\[\(.*\)\]$/\1/')
	echo "doudoulinux-sound: setting default card $SELECTCARD"
	
	# set card
	set_audiocard "$CARDNBR" "$CARDNAME"
}
